from dnfpluginscore import _, logger

import binascii
import itertools
import libpkgmanifest
import os
import shutil

import dnf
import dnf.cli
import dnf.util
import hawkey
import rpm


DEFAULT_MANIFEST_FILENAME = 'packages.manifest.yaml'
MODULE_FILENAME = 'modules_dump.modulemd.yaml'
MODULAR_DATA_SEPARATOR = '...'


@dnf.plugin.register_command
class ManifestCommand(dnf.cli.Command):
    aliases = ("manifest",)
    summary = _("Operations for working with RPM package manifest files")

    def __init__(self, cli):
        super(ManifestCommand, self).__init__(cli)
        self.cmd = None
        self.file = None
        self.download_dir = None
        self.module_packages = []

        # keep reference to the manifest so inner objects are not destroyed when going out of scope
        self.parsed_manifest = None

    @staticmethod
    def set_argparser(parser):
        parser.add_argument("subcommand", nargs=1, choices=['new', 'download', 'install'])
        parser.add_argument('specs', nargs='*', help=_('package specs to be processed'))
        parser.add_argument('--file', help=_('manifest file path to use'))
        parser.add_argument("--source", action='store_true', help=_('include also source packages'))
        parser.add_argument("--url-only", action='store_true', help=_('use package URLs from manifest for downloading'))

    def configure(self):
        self.cmd = self.opts.subcommand[0]

        repos_needed = not (self.cmd == 'download' and self.opts.url_only)
        only_system_repo = self.cmd == 'new' and not self.opts.specs

        demands = self.cli.demands

        if repos_needed:
            demands.sack_activation = True
            if not only_system_repo:
                demands.available_repos = True

        self.base.conf.strict = True

        if self.cmd == 'install':
            demands.resolving = True

        if self.opts.source:
            self.base.repos.enable_source_repos()

        if self.opts.file:
            self.file = self.opts.file
        else:
            self.file = DEFAULT_MANIFEST_FILENAME

        if self.opts.destdir:
            self.download_dir = self.opts.destdir
        else:
            self.download_dir, _ = os.path.splitext(os.path.join(dnf.i18n.ucd(os.getcwd()), self.file))

        if self.cmd in ['install', 'download']:
            self.base.conf.destdir = self.download_dir

    def run(self):
        match self.cmd:
            case 'new':
                self._new()
            case 'download':
                self._download()
            case 'install':
                self._install()

    def _new(self):
        """
        Generate a new manifest file based on the provided package specs.

        The specs are resolved, and all packages and their dependencies
        are recorder in the manifest file, considering the current system
        state and dnf configuration.

        If no specs are provided, the manifest is generated from the
        installed packages on the system.
        """
        self._generate(self.opts.specs)

    def _download(self):
        """
        Download all packages specified in the manifest file to disk.
        """
        if self.opts.url_only:
            for path in self._manifest_to_pkg_urls():
                tmp_path = dnf.util._urlopen_progress(path, self.base.conf, self.base.output.progress)
                shutil.move(tmp_path, self.download_dir)
        else:
            specs = self._manifest_to_pkg_specs()
            pkgs = self._get_packages(specs)
            self.base.download_packages(pkgs, self.base.output.progress)

        self._dump_modular_data()

    def _install(self):
        """
        Install all packages specified in the manifest file.

        Packages previously downloaded with the 'download' subcommand
        are reused for the transaction.

        Throws a dnf exception if the state of the packages in the manifest
        does not match the current state of the repositories.
        """
        for spec in self._manifest_to_pkg_specs():
            self.base.install(spec)

    def _generate(self, specs):
        self._prepare_modular_data()
        modules_info = {}

        if self.opts.specs:
            dnf_pkgs = self._get_packages_with_deps(specs)
        else:
            dnf_pkgs = self.base.sack.query()

        manifest = libpkgmanifest.Manifest()

        for dnf_pkg in dnf_pkgs:
            pkg = libpkgmanifest.Package()
            pkg.arch = dnf_pkg.arch
            pkg.nevra = self._get_package_nevra(dnf_pkg)
            pkg.size = dnf_pkg.size

            if dnf_pkg._is_local_pkg():
                pkg.repo_id = dnf_pkg.from_repo
            else:
                pkg.repo_id = dnf_pkg.repoid
                pkg.url = dnf_pkg.remote_location()

            if dnf_pkg.arch != 'src':
                pkg.srpm = dnf.util.rtrim(dnf_pkg.sourcerpm, ".rpm")

            modular_pkg = self._retrieve_modular_pkg(dnf_pkg)
            if modular_pkg:
                pkg.module.name = modular_pkg.getName()
                pkg.module.stream = modular_pkg.getStream()
                modules_info[modular_pkg.getId()] = modular_pkg.getYaml()

            pkg.checksum.method, pkg.checksum.digest = self._retrieve_pkg_checksum(dnf_pkg)

            manifest.packages.add(pkg)

        serializer = libpkgmanifest.Serializer()
        serializer.serialize(manifest, self.file)

        # append modular yaml data to the manifest file
        if modules_info:
            with open(self.file, "a") as f:
                f.write('\n' + MODULAR_DATA_SEPARATOR + '\n')
                for module in modules_info.values():
                    f.write(module)

    def _manifest_to_pkgs(self):
        if not os.path.isfile(self.file):
            raise dnf.exceptions.Error(_("Input manifest file '%s' does not exist") % self.file)

        manifest = libpkgmanifest.Parser().parse(self.file)
        self.parsed_manifest = manifest
        pkgs = [pkg for pkgs in manifest.packages.get().values() for pkg in pkgs]
        if self.cmd == 'install' or not self.opts.source:
            pkgs = [pkg for pkg in pkgs if pkg.arch != 'src']
        return pkgs

    def _manifest_to_pkg_specs(self):
        return [pkg.nevra for pkg in self._manifest_to_pkgs()]

    def _manifest_to_pkg_urls(self):
        return [pkg.url for pkg in self._manifest_to_pkgs()]

    def _get_packages_with_deps(self, pkg_specs):
        pkgs = set(self._get_packages(pkg_specs))
        if self.cmd == 'generate' and self.opts.source:
            pkgs |= set(self._get_packages(pkg_specs, True))

        pkg_set = set(pkgs)
        for pkg in pkgs:
            goal = hawkey.Goal(self.base.sack)
            goal.install(pkg)
            rc = goal.run(ignore_weak_deps=(not self.base.conf.install_weak_deps))
            if rc:
                pkg_set.update(goal.list_installs())
                pkg_set.update(goal.list_upgrades())
            else:
                msg = [_('Error in resolve of packages:')]
                logger.error("\n    ".join(msg + [str(pkg) for pkg in pkgs]))
                logger.error(dnf.util._format_resolve_problems(goal.problem_rules()))
                raise dnf.exceptions.Error()
        return pkg_set

    def _get_packages(self, pkg_specs, source=False):
        func = self._get_query_source if source else self._get_query
        queries = []
        for pkg_spec in pkg_specs:
            try:
                queries.append(func(pkg_spec))
            except dnf.exceptions.PackageNotFoundError as e:
                logger.error(dnf.i18n.ucd(e))
                raise dnf.exceptions.Error(e)

        pkgs = list(itertools.chain(*queries))
        return pkgs

    def _get_query(self, pkg_spec):
        subj = dnf.subject.Subject(pkg_spec)
        q = subj.get_best_query(self.base.sack, with_src=self.opts.source)
        q = q.available()
        q = q.filterm(latest_per_arch_by_priority=True)
        if len(q.run()) == 0:
            msg = _("No package %s available.") % (pkg_spec)
            raise dnf.exceptions.PackageNotFoundError(msg)
        return q

    def _get_query_source(self, pkg_spec):
        subj = dnf.subject.Subject(pkg_spec)
        for nevra_obj in subj.get_nevra_possibilities():
            tmp_query = nevra_obj.to_query(self.base.sack).available()
            if tmp_query:
                return tmp_query.latest()

        msg = _("No package %s available.") % (pkg_spec)
        raise dnf.exceptions.PackageNotFoundError(msg)

    def _prepare_modular_data(self):
        if dnf.base.WITH_MODULES:
            self.module_packages = self.base._moduleContainer.getModulePackages()

    def _retrieve_modular_pkg(self, pkg):
        for module_pkg in self.module_packages:
            if self._get_package_nevra(pkg) in module_pkg.getArtifacts():
                return module_pkg
        return None

    def _dump_modular_data(self):
        module_found = False
        with open(self.file, 'r') as infile:
            for line in infile:
                if MODULAR_DATA_SEPARATOR in line:
                    module_found = True
                    break
            if module_found:
                with open(os.path.join(self.download_dir, MODULE_FILENAME), 'w') as outfile:
                    for line in infile:
                        outfile.write(line)

    def _retrieve_pkg_checksum(self, pkg):
        if pkg._is_local_pkg():
            hdr = self._retrieve_rpm_hdr(pkg)
            method = self._rpm_checksum_type_to_manifest_conversion(hdr[rpm.RPMTAG_PAYLOADDIGESTALGO])
            digest = hdr[rpm.RPMTAG_PAYLOADDIGEST][0]
        else:
            dnf_chksum_type, dnf_chksum_digest = pkg.chksum
            method = self._dnf_checksum_type_to_manifest_conversion(dnf_chksum_type)
            digest = binascii.hexlify(dnf_chksum_digest).decode()
        return method, digest

    def _retrieve_rpm_hdr(self, dnf_pkg):
        ts = rpm.TransactionSet()
        mi = ts.dbMatch('name', dnf_pkg.name)
        for hdr in mi:
            if hdr['version'] == dnf_pkg.version and \
               hdr['release'] == dnf_pkg.release and \
               hdr['arch'] == dnf_pkg.arch:
               return hdr
        return None

    def _rpm_checksum_type_to_manifest_conversion(self, checksum_type):
        match checksum_type:
            case 1:
                return libpkgmanifest.ChecksumMethod_MD5
            case 2:
                return libpkgmanifest.ChecksumMethod_SHA1
            case 8:
                return libpkgmanifest.ChecksumMethod_SHA256
            case 9:
                return libpkgmanifest.ChecksumMethod_SHA384
            case 10:
                return libpkgmanifest.ChecksumMethod_SHA512
            case _:
                raise dnf.exceptions.Error(_("Unknown RPM package checksum type: %s") % checksum_type)

    def _dnf_checksum_type_to_manifest_conversion(self, checksum_type):
        match checksum_type:
            case hawkey.CHKSUM_MD5:
                return libpkgmanifest.ChecksumMethod_MD5
            case hawkey.CHKSUM_SHA1:
                return libpkgmanifest.ChecksumMethod_SHA1
            case hawkey.CHKSUM_SHA256:
                return libpkgmanifest.ChecksumMethod_SHA256
            case hawkey.CHKSUM_SHA384:
                return libpkgmanifest.ChecksumMethod_SHA384
            case hawkey.CHKSUM_SHA512:
                return libpkgmanifest.ChecksumMethod_SHA512
            case _:
                raise dnf.exceptions.Error(_("Unknown DNF package checksum type: %s") % checksum_type)

    def _get_package_nevra(self, pkg):
        name = pkg.name
        epoch = pkg.epoch if pkg.epoch is not None else "0"
        version = pkg.version
        release = pkg.release
        arch = pkg.arch

        return f"{name}-{epoch}:{version}-{release}.{arch}"
